@@ -1,11 +1,12 @@
 /* wdiff -- front end to diff for comparing on a word per word basis.
-   Copyright (C) 1992, 1997, 1998, 1999 Free Software Foundation, Inc.
+   Copyright (C) 1992, 1997, 1998, 1999, 2009,
+                 2010 Free Software Foundation, Inc.
    Francois Pinard <pinard@iro.umontreal.ca>, 1992.
 
-   This program is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -13,8 +14,8 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 #include "system.h"
 
@@ -82,12 +83,10 @@ const char *tgetstr ();
 #endif
 
 #include <sys/stat.h>
-
-#if HAVE_UNISTD_H
-# include <unistd.h>
-#endif
-
+#include <unistd.h>
 #include <getopt.h>
+#include <locale.h>
+#include <sys/wait.h>
 
 char *getenv ();
 FILE *readpipe PARAMS ((const char *, ...));
@@ -117,6 +116,7 @@ struct option const longopts[] =
   {"statistics"  , 0, NULL, 's'},
   {"terminal"    , 0, NULL, 't'},
   {"version"     , 0, NULL, 'v'},
+  {"diff-input"  , 0, NULL, 'd'},
   {NULL          , 0, NULL, 0}
 };
 
@@ -125,6 +125,7 @@ const char *program_name;	/* name of executing program */
 int inhibit_left;		/* inhibit display of left side words */
 int inhibit_right;		/* inhibit display of left side words */
 int inhibit_common;		/* inhibit display of common words */
+int diff_input;			/* expect (unified) diff as input */
 int ignore_case;		/* ignore case in comparisons */
 int show_statistics;		/* if printing summary statistics */
 int no_wrapping;		/* end/restart strings at end of lines */
@@ -184,9 +185,9 @@ int character;			/* for reading input_file */
 char directive;			/* diff directive character */
 int argument[4];		/* four diff directive arguments */
 
-FILE *output_file;		/* file to which we write output */
-const char *termcap_init_string; /* how to initialize the termcap mode */
-const char *termcap_end_string; /* how to complete the termcap mode */
+FILE *output_file;			/* file to which we write output */
+const char *termcap_init_string;	/* how to initialize the termcap mode */
+const char *termcap_end_string;		/* how to complete the termcap mode */
 
 int count_total_left;		/* count of total words in left file */
 int count_total_right;		/* count of total words in right file */
@@ -223,8 +224,8 @@ setup_signals (void)
   signal (SIGPIPE, signal_handler);
   signal (SIGTERM, signal_handler);
 }
-
 
+
 /* Terminal initialization.  */
 
 static void
@@ -241,13 +242,13 @@ initialize_strings (void)
 
       name = getenv ("TERM");
       if (name == NULL)
-	error (1, 0,
-	       _("Select a terminal through the TERM environment variable."));
+	error (EXIT_FAILURE, 0,
+	       _("select a terminal through the TERM environment variable"));
       success = tgetent (term_buffer, name);
       if (success < 0)
-	error (1, 0, _("Could not access the termcap data base."));
+	error (EXIT_FAILURE, 0, _("could not access the termcap data base"));
       if (success == 0)
-	error (1, 0, _("Terminal type `%s' is not defined."), name);
+	error (EXIT_FAILURE, 0, _("terminal type `%s' is not defined"), name);
       buffer = (char *) malloc (strlen (term_buffer));
       filler = buffer;
 
@@ -282,8 +283,8 @@ initialize_strings (void)
 	user_insert_end = "+}";
     }
 }
-
 
+
 /* Character input and output.  */
 
 #if HAVE_TPUTS
@@ -431,10 +432,10 @@ copy_whitespace (SIDE *side, FILE *file)
     {
 
       /* While changing lines, ensure we stop any special display prior
-	 to, and restore the special display after.  When copy_mode is
-	 anything else than COPY_NORMAL, file is always output_file.  We
-	 care underlining whitespace or overstriking it with itself,
-	 because "less" understands these things as emphasis requests.  */
+         to, and restore the special display after.  When copy_mode is
+         anything else than COPY_NORMAL, file is always output_file.  We
+         care underlining whitespace or overstriking it with itself,
+         because "less" understands these things as emphasis requests.  */
 
       switch (copy_mode)
 	{
@@ -508,7 +509,7 @@ copy_whitespace (SIDE *side, FILE *file)
 `-----------------------------------------*/
 
 static void
-copy_word (SIDE *side, FILE *file)
+copy_word (SIDE * side, FILE * file)
 {
   if (interrupted)
     longjmp (signal_label, 1);
@@ -517,7 +518,7 @@ copy_word (SIDE *side, FILE *file)
     {
 
       /* In printer mode, act according to copy_mode.  If copy_mode is not
-	 COPY_NORMAL, we know that file is necessarily output_file.  */
+         COPY_NORMAL, we know that file is necessarily output_file.  */
 
       if (overstrike)
 	switch (copy_mode)
@@ -562,7 +563,7 @@ static char *
 create_template_filename (char *tmpl, size_t tmpl_len)
 {
   struct stat stat_buffer;	/* for checking if file is directory */
-  const char* dir;
+  const char *dir;
   size_t dirlen;
 
   dir = getenv ("TMPDIR");
@@ -592,64 +593,143 @@ create_template_filename (char *tmpl, size_t tmpl_len)
       return NULL;
     }
 
-  sprintf(tmpl, "%.*s/wdiff.XXXXXX", (int) dirlen, dir);
+  sprintf (tmpl, "%.*s/wdiff.XXXXXX", (int) dirlen, dir);
   return tmpl;
 }
 
+/*--------------------------------------------------.
+| Create a temporary file for one side of the diff. |
+`--------------------------------------------------*/
+static void
+create_temporary_side (SIDE * side)
+{
+  int fd;	/* for file descriptors returned by mkstemp */
+
+  /* Select a file name, use it for opening a temporary file and
+     unlink it right away. We do not need the file name itself
+     anymore.  */
+
+  if (create_template_filename (side->temp_name, L_tmpnam) == NULL)
+    error (EXIT_FAILURE, 0, _("no suitable temporary directory exists"));
+  if ((fd = mkstemp (side->temp_name)) == -1)
+    error (EXIT_FAILURE, errno, "%s", side->temp_name);
+
+  side->file = fdopen (fd, "w+");
+  if (side->file == NULL)
+    error (EXIT_FAILURE, errno, "%s", side->temp_name);
+  if (unlink (side->temp_name) != 0)
+    error (EXIT_FAILURE, errno, "%s", side->temp_name);
+}
+
+/*--------------------------------------------------------.
+| Read unified diff and produce two output files from it. |
+`--------------------------------------------------------*/
+
+static void
+split_diff (const char *path)
+{
+  FILE *input;
+  int character;
+  int start_of_line = 1;
+  int output_to = 3;
+
+  if (path == NULL)
+    {
+      input = stdin;
+    }
+  else
+    {
+      input = fopen (path, "r");
+      if (input == NULL)
+	error (EXIT_FAILURE, errno, "%s", path);
+    }
+
+  create_temporary_side (left_side);
+  create_temporary_side (right_side);
+
+  while ((character = getc (input)) != EOF)
+    {
+      if (start_of_line)
+	{
+	  start_of_line = 0;
+	  switch (character)
+	    {
+	    case '-':
+	      output_to = 1;
+	      continue;
+	    case '+':
+	      output_to = 2;
+	      continue;
+	    case ' ':
+	      output_to = 3;
+	      continue;
+	    default:
+	      output_to = 3;
+	      break;
+	    }
+	}
+      if (output_to & 1)
+	putc (character, left_side->file);
+      if (output_to & 2)
+	putc (character, right_side->file);
+      start_of_line = (character == '\n' || character == '\r');
+    }
+  rewind (left_side->file);
+  rewind (right_side->file);
+}
+
 /*-------------------------------------------------------------------------.
 | For a given SIDE, turn original input file in another one, in which each |
 | word is on one line.							   |
 `-------------------------------------------------------------------------*/
 
 static void
-split_file_into_words (SIDE *side)
+split_file_into_words (SIDE * side)
 {
   struct stat stat_buffer;	/* for checking if file is directory */
-  int fd;                /* for file descriptors returned by mkstemp */
+  int fd;			/* for file descriptors returned by mkstemp */
+  FILE *input;			/* used when copying from non-seekable input */
 
   /* Open files.  */
 
-  if (side->filename == NULL)
+  if (!diff_input)
     {
+      if (side->filename == NULL)
+	{
+	  side->file = stdin;
+	}
+      else
+	{
+	  /* Check and diagnose if the file name is a directory.  Or else,
+	     prepare the file for reading.  */
 
-      /* Select a file name, use it for opening a temporary file and
-	 unlink it right away.  Then, copy the whole standard input on
-	 this temporary local file.  Once done, prepare it for reading.
-	 We do not need the file name itself anymore.  */
+	  if (stat (side->filename, &stat_buffer) != 0)
+	    error (EXIT_FAILURE, errno, "%s", side->filename);
+	  if ((stat_buffer.st_mode & S_IFMT) == S_IFDIR)
+	    error (EXIT_FAILURE, 0, _("directories not supported"));
+	  side->file = fopen (side->filename, "r");
+	  if (side->file == NULL)
+	    error (EXIT_FAILURE, errno, "%s", side->filename);
+	}
 
-      if (create_template_filename (side->temp_name, L_tmpnam) == NULL)
-	error (EXIT_FAILURE, 0, _("No suitable temporary directory exists"));
-      if ((fd = mkstemp (side->temp_name)) == -1)
-        error (EXIT_FAILURE, errno, "%s", side->temp_name);
+      if (fseek (side->file, 0L, SEEK_CUR) != 0)
+	{
+	  /* Non-seekable input, e.g. stdin or shell process substitution.
+	     Copy the whole input to a temporary local file.  Once done,
+	     prepare it for reading.  */
+	  input = side->file;
+	  create_temporary_side (side);
+	  while (side->character = getc (input), side->character != EOF)
+	    putc (side->character, side->file);
+	  rewind (side->file);
 
-      side->file = fdopen (fd, "w+");
-      if (side->file == NULL)
-	error (EXIT_FAILURE, errno, "%s", side->temp_name);
-      if (unlink (side->temp_name) != 0)
-	error (EXIT_FAILURE, errno, "%s", side->temp_name);
-      while (side->character = getchar (), side->character != EOF)
-	putc (side->character, side->file);
-      rewind (side->file);
-    }
-  else
-    {
-
-      /* Check and diagnose if the file name is a directory.  Or else,
-	 prepare the file for reading.  */
-
-      if (stat (side->filename, &stat_buffer) != 0)
-	error (EXIT_FAILURE, errno, "%s", side->filename);
-      if ((stat_buffer.st_mode & S_IFMT) == S_IFDIR)
-	error (EXIT_FAILURE, 0, _("Directories not supported"));
-      side->file = fopen (side->filename, "r");
-      if (side->file == NULL)
-	error (EXIT_FAILURE, errno, "%s", side->filename);
+	}
     }
   side->character = getc (side->file);
   side->position = 0;
 
   if (create_template_filename (side->temp_name, L_tmpnam) == NULL)
-    error (EXIT_FAILURE, 0, _("No suitable temporary directory exists"));
+    error (EXIT_FAILURE, 0, _("no suitable temporary directory exists"));
   if ((fd = mkstemp (side->temp_name)) == -1)
     error (EXIT_FAILURE, errno, "%s", side->temp_name);
 
@@ -675,9 +755,9 @@ split_file_into_words (SIDE *side)
 
 /*-------------------------------------------------------------------.
 | Decode one directive line from INPUT_FILE.  The format should be:  |
-| 								     |
-|      ARG0 [ , ARG1 ] LETTER ARG2 [ , ARG3 ] \n		     |
-| 								     |
+|                                                                    |
+|      ARG0 [ , ARG1 ] LETTER ARG2 [ , ARG3 ] \n                     |
+|                                                                    |
 | By default, ARG1 is assumed to have the value of ARG0, and ARG3 is |
 | assumed to have the value of ARG2.  Return 0 if any error found.   |
 `-------------------------------------------------------------------*/
@@ -756,7 +836,7 @@ decode_directive_line (void)
 `----------------------------------------------*/
 
 static void
-skip_until_ordinal (SIDE *side, int ordinal)
+skip_until_ordinal (SIDE * side, int ordinal)
 {
   while (side->position < ordinal)
     {
@@ -770,7 +850,7 @@ skip_until_ordinal (SIDE *side, int ordinal)
 `----------------------------------------------*/
 
 static void
-copy_until_ordinal (SIDE *side, int ordinal)
+copy_until_ordinal (SIDE * side, int ordinal)
 {
   while (side->position < ordinal)
     {
@@ -795,7 +875,7 @@ reformat_diff_output (void)
   left_side->character = getc (left_side->file);
   left_side->position = 0;
 
-  rewind(right_side->file);
+  rewind (right_side->file);
   right_side->character = getc (right_side->file);
   right_side->position = 0;
 
@@ -859,16 +939,20 @@ reformat_diff_output (void)
 	    }
 
 	  if (!inhibit_left)
-	    if (!inhibit_common && inhibit_right)
-	      copy_until_ordinal (left_side, resync_left);
-	    else
-	      skip_until_ordinal (left_side, resync_left);
+	    {
+	      if (!inhibit_common && inhibit_right)
+		copy_until_ordinal (left_side, resync_left);
+	      else
+		skip_until_ordinal (left_side, resync_left);
+	    }
 
 	  if (!inhibit_right)
-	    if (inhibit_common)
-	      skip_until_ordinal (right_side, resync_right);
-	    else
-	      copy_until_ordinal (right_side, resync_right);
+	    {
+	      if (inhibit_common)
+		skip_until_ordinal (right_side, resync_right);
+	      else
+		copy_until_ordinal (right_side, resync_right);
+	    }
 
 	  if (!inhibit_common && inhibit_left && inhibit_right)
 	    copy_until_ordinal (right_side, resync_right);
@@ -938,8 +1022,8 @@ reformat_diff_output (void)
   fclose (left_side->file);
   fclose (right_side->file);
 }
-
 
+
 /* Launch and complete various programs.  */
 
 /*-------------------------.
@@ -1131,8 +1215,8 @@ print_statistics (void)
     }
   printf ("\n");
 }
-
 
+
 /* Main control.  */
 
 /*-----------------------------------.
@@ -1143,10 +1227,10 @@ static void
 print_copyright (void)
 {
   fputs (_("\
-This program is free software; you can redistribute it and/or modify\n\
+This program is free software: you can redistribute it and/or modify\n\
 it under the terms of the GNU General Public License as published by\n\
-the Free Software Foundation; either version 2, or (at your option)\n\
-any later version.\n\
+the Free Software Foundation, either version 3 of the License, or\n\
+(at your option) any later version.\n\
 \n\
 This program is distributed in the hope that it will be useful,\n\
 but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
@@ -1154,9 +1238,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
 GNU General Public License for more details.\n\
 \n\
 You should have received a copy of the GNU General Public License\n\
-along with this program; if not, write to the Free Software Foundation,\n\
-Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.\n"),
-	 stdout);
+along with this program.  If not, see <http://www.gnu.org/licenses/>.\n"),
+         stdout);
 }
 
 /*-----------------------------.
@@ -1174,37 +1257,36 @@ usage (int status)
       fputs (_("\
 wdiff - Compares words in two files and report differences.\n"),
 	     stdout);
+      fputs ("\n", stdout);
       printf (_("\
-\n\
-Usage: %s [OPTION]... FILE1 FILE2\n"),
-	      program_name);
-      printf (_("\
-Mandatory arguments to long options are mandatory for short options too.\n\
-\n\
-  -C, --copyright            print copyright then exit\n\
-  -K, --no-init-term         like -t, but no termcap init/term strings\n\
-  -1, --no-deleted           inhibit output of deleted words\n\
-  -2, --no-inserted          inhibit output of inserted words\n\
-  -3, --no-common            inhibit output of common words\n"));
-      printf (_("\
-  -a, --auto-pager           automatically calls a pager\n\
-  -h, --help                 print this help\n\
-  -i, --ignore-case          fold character case while comparing\n\
-  -l, --less-mode            variation of printer mode for \"less\"\n\
-  -n, --avoid-wraps          do not extend fields through newlines\n\
-  -p, --printer              overstrike as for printers\n"));
-      printf (_("\
-  -s, --statistics           say how many words deleted, inserted etc.\n\
-  -t, --terminal             use termcap as for terminal displays\n\
-  -v, --version              print program version then exit\n\
-  -w, --start-delete=STRING  string to mark beginning of delete region\n\
-  -x, --end-delete=STRING    string to mark end of delete region\n\
-  -y, --start-insert=STRING  string to mark beginning of insert region\n\
-  -z, --end-insert=STRING    string to mark end of insert region\n"));
+Usage: %s [OPTION]... FILE1 FILE2\n\
+       %s -d [OPTION]... [FILE]\n"),
+	      program_name, program_name);
+      fputs ("\n", stdout);
       fputs (_("\
-\n\
-Report bugs to <wdiff-bugs@gnu.org>.\n"),
-	     stdout);
+Mandatory arguments to long options are mandatory for short options too.\n"),
+             stdout);
+      fputs (_("  -C, --copyright            display copyright then exit\n"), stdout);
+      fputs (_("  -K, --no-init-term         like -t, but no termcap init/term strings\n"), stdout);
+      fputs (_("  -1, --no-deleted           inhibit output of deleted words\n"), stdout);
+      fputs (_("  -2, --no-inserted          inhibit output of inserted words\n"), stdout);
+      fputs (_("  -3, --no-common            inhibit output of common words\n"), stdout);
+      fputs (_("  -a, --auto-pager           automatically calls a pager\n"), stdout);
+      fputs (_("  -d, --diff-input           use single unified diff as input\n"), stdout);
+      fputs (_("  -h, --help                 display this help then exit\n"), stdout);
+      fputs (_("  -i, --ignore-case          fold character case while comparing\n"), stdout);
+      fputs (_("  -l, --less-mode            variation of printer mode for \"less\"\n"), stdout);
+      fputs (_("  -n, --avoid-wraps          do not extend fields through newlines\n"), stdout);
+      fputs (_("  -p, --printer              overstrike as for printers\n"), stdout);
+      fputs (_("  -s, --statistics           say how many words deleted, inserted etc.\n"), stdout);
+      fputs (_("  -t, --terminal             use termcap as for terminal displays\n"), stdout);
+      fputs (_("  -v, --version              display program version then exit\n"), stdout);
+      fputs (_("  -w, --start-delete=STRING  string to mark beginning of delete region\n"), stdout);
+      fputs (_("  -x, --end-delete=STRING    string to mark end of delete region\n"), stdout);
+      fputs (_("  -y, --start-insert=STRING  string to mark beginning of insert region\n"), stdout);
+      fputs (_("  -z, --end-insert=STRING    string to mark end of insert region\n"), stdout);
+      fputs ("\n", stdout);
+      fputs (_("Report bugs to <wdiff-bugs@gnu.org>.\n"), stdout);
     }
   exit (status);
 }
@@ -1229,6 +1311,7 @@ main (int argc, char *const argv[])
   inhibit_right = 0;
   inhibit_common = 0;
 
+  diff_input = 0;
   ignore_case = 0;
   show_statistics = 0;
   no_wrapping = 0;
@@ -1255,7 +1338,7 @@ main (int argc, char *const argv[])
   count_changed_right = 0;
 
   while (option_char = getopt_long (argc, (char **) argv,
-				    "123CKahilnpstvw:x:y:z:", longopts, NULL),
+				    "123CKadhilnpstvw:x:y:z:", longopts, NULL),
 	 option_char != EOF)
     switch (option_char)
       {
@@ -1279,6 +1362,10 @@ main (int argc, char *const argv[])
 	autopager = 1;
 	break;
 
+      case 'd':
+	diff_input = 1;
+	break;
+
       case 'h':
 	usage (EXIT_SUCCESS);
 
@@ -1315,15 +1402,15 @@ main (int argc, char *const argv[])
 	  find_termcap = 1;
 	break;
 #else
-	error (EXIT_FAILURE, 0,
-	       _("Cannot use -t, termcap not available."));
+	error (EXIT_FAILURE, 0, _("cannot use -t, termcap not available"));
 #endif
 
       case 'v':
 	printf ("wdiff (GNU %s) %s\n", PACKAGE, VERSION);
 	fputs (_("\
 \n\
-Copyright (C) 1992, 1997 Free Software Foundation, Inc.\n"),
+Copyright (C) 1992, 1997, 1998, 1999, 2009, 2010 Free Software\n\
+Foundation, Inc.\n"),
 	     stdout);
 	fputs (_("\
 This is free software; see the source for copying conditions.  There is NO\n\
@@ -1355,12 +1442,6 @@ Written by Franc,ois Pinard <pinard@iro.umontreal.ca>.\n"),
 	usage (EXIT_FAILURE);
       }
 
-  if (optind + 2 != argc)
-    {
-      error (0, 0, _("Missing file arguments"));
-      usage (EXIT_FAILURE);
-    }
-
   /* If find_termcap still undecided, make it true only if autopager is set
      while stdout is directed to a terminal.  This decision might be
      reversed later, if the pager happens to be "less".  */
@@ -1370,22 +1451,49 @@ Written by Franc,ois Pinard <pinard@iro.umontreal.ca>.\n"),
 
   /* Setup file names and signals, then do it all.  */
 
-  if (strcmp (argv[optind], "") == 0 || strcmp (argv[optind], "-") == 0)
-    left_side->filename = NULL;
+  if (diff_input)
+    {
+      if (optind + 1 < argc)
+	{
+	  error (0, 0, _("too many file arguments"));
+	  usage (EXIT_FAILURE);
+	}
+      if (optind == argc || strcmp (argv[optind], "") == 0
+	  || strcmp (argv[optind], "-") == 0)
+	split_diff (NULL);
+      else
+	split_diff (argv[optind]);
+    }
   else
-    left_side->filename = argv[optind];
-  optind++;
-  *left_side->temp_name = '\0';
+    {
+      if (optind + 2 > argc)
+	{
+	  error (0, 0, _("missing file arguments"));
+	  usage (EXIT_FAILURE);
+	}
+      if (optind + 2 < argc)
+	{
+	  error (0, 0, _("too many file arguments"));
+	  usage (EXIT_FAILURE);
+	}
 
-  if (strcmp (argv[optind], "") == 0 || strcmp (argv[optind], "-") == 0)
-    right_side->filename = NULL;
-  else
-    right_side->filename = argv[optind];
-  optind++;
-  *right_side->temp_name = '\0';
+      if (strcmp (argv[optind], "") == 0 || strcmp (argv[optind], "-") == 0)
+	left_side->filename = NULL;
+      else
+	left_side->filename = argv[optind];
+      optind++;
+      *left_side->temp_name = '\0';
 
-  if (left_side->filename == NULL && right_side->filename == NULL)
-    error (EXIT_FAILURE, 0, _("Only one file may be standard input."));
+      if (strcmp (argv[optind], "") == 0 || strcmp (argv[optind], "-") == 0)
+	right_side->filename = NULL;
+      else
+	right_side->filename = argv[optind];
+      optind++;
+      *right_side->temp_name = '\0';
+
+      if (left_side->filename == NULL && right_side->filename == NULL)
+	error (EXIT_FAILURE, 0, _("only one file may be standard input"));
+    }
 
   setup_signals ();
   input_file = NULL;
